def manage_addZMySQLConnection(self, id, title,
                                connection_string,
                                check=None, REQUEST=None):
    """Add a MySQL connection to a folder.

    Arguments:

        REQUEST -- The current request

        title -- The title of the ZMySQLDA Connection (string)

        id -- The id of the ZMySQLDA Connection (string)

        connection_string -- The connection string is of the form:

                'database[@host[:port]] [user [password [unix_socket]]]'

        or typically:

                'database user password'

        to use a MySQL server on localhost via the standard UNIX socket.
        Only specify host if the server is on a remote system. You can
        use a non-standard port, if necessary. If the UNIX socket is in
        a non-standard location, you can specify the full path to it
        after the password.

    """

class Connection:
    """MySQL Connection Object"""

    __constructor__ = manage_addZMySQLConnection





import transaction

from Shared.DC.ZRDB.TM import TM
from ZODB.POSException import TransactionFailedError

import logging
LOG = logging.getLogger('ZMySQLDA')

class joinTM(TM):

    _registered = False
    _finalize   = False

    def _register(self):
        """ Override to replace transaction register() call with join().

            The primary reason for this is to eliminate issues in both the
            super's _register() and transaction register(). The former has a
            bare except: that hides useful errors. The latter deals poorly with
            exceptions raised from the join due to state modifications made
            before the join attempt.
    
            They also both (for our purposes) needlessly add wrapper objects
            around self, resulting in unnecessary overhead.
        """
        if not self._registered:
            try:
                transaction.get().join(self)
            except TransactionFailedError:
                msg = "database connection failed to join transaction."
                LOG.error(msg)
            except ValueError:
                msg = "database connection failed to join transaction."
                LOG.error(msg, exc_info=True)
                raise
            else:
                self._begin()
                self._registered = True
                self._finalize   = False





##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

'''$Id: db.py 649 2010-11-10 02:27:54Z eikenberry $'''
__version__='$Revision: 649 $'[11:-2]

import time
import _mysql
import MySQLdb
from _mysql_exceptions import (OperationalError, NotSupportedError,
        ProgrammingError)
MySQLdb_version_required = (1,2,1)

try:
    MySQLdb_version = MySQLdb.version_info[:3]
except AttributeError:
    MySQLdb_version = getattr(_mysql, 'version_info', (0,0,0))

if MySQLdb_version < MySQLdb_version_required:
    raise NotSupportedError, \
        "ZMySQLDA requires at least MySQLdb %s, %s found" % \
        (MySQLdb_version_required, _v)

from MySQLdb.converters import conversions
from MySQLdb.constants import FIELD_TYPE, CR, ER, CLIENT, FLAG
from ZODB.POSException import ConflictError
from DateTime import DateTime

from thread import get_ident, allocate_lock
import logging
LOG = logging.getLogger('ZMySQLDA')
from joinTM import joinTM

hosed_connection = {
        CR.SERVER_GONE_ERROR:    "Server gone.",
        CR.SERVER_LOST:          "Server lost.",
        CR.COMMANDS_OUT_OF_SYNC: ("Commands out of sync. Possibly a misplaced"
            " semicolon (;) in a query."),
    }

query_syntax_error = (
    ER.BAD_FIELD_ERROR,
    )

key_types = {
    "PRI": "PRIMARY KEY",
    "MUL": "INDEX",
    "UNI": "UNIQUE",
    }

field_icons = "bin", "date", "datetime", "float", "int", "text", "time"

icon_xlate = {
    "varchar": "text", "char": "text",
    "enum": "what", "set": "what",
    "double": "float", "numeric": "float",
    "blob": "bin", "mediumblob": "bin", "longblob": "bin",
    "tinytext": "text", "mediumtext": "text",
    "longtext": "text", "timestamp": "datetime",
    "decimal": "float", "smallint": "int",
    "mediumint": "int", "bigint": "int",
    }

type_xlate = {
    "double": "float", "numeric": "float",
    "decimal": "float", "smallint": "int",
    "mediumint": "int", "bigint": "int",
    "int": "int", "float": "float",
    "timestamp": "datetime", "datetime": "datetime",
    "time": "datetime",
    }

def _mysql_timestamp_converter(s):
    if len(s) < 14:
        s = s + "0"*(14-len(s))
        parts = map(int, (s[:4],s[4:6],s[6:8],
                          s[8:10],s[10:12],s[12:14]))
    return DateTime("%04d-%02d-%02d %02d:%02d:%02d" % tuple(parts))

def DateTime_or_None(s):
    try: return DateTime(s)
    except: return None

def int_or_long(s):
    try: return int(s)
    except: return long(s)


class DBPool(object):
    """
      This class is an interface to DB.
      Its caracteristic is that an instance of this class interfaces multiple
      instanes of DB class, each one being bound to a specific thread.
    """

    connected_timestamp = ''
    _create_db          = False
    use_unicode         = False

    def __init__(self, db_cls, create_db=False, use_unicode=False):
        """ Set transaction managed DB class for use in pool.
        """
        self.DB = db_cls
        # pool of one db object/thread
        self._db_pool = {}
        self._db_lock = allocate_lock()
        # auto-create db if not present on server
        self._create_db = create_db
        # unicode settings
        self.use_unicode = use_unicode

    def __call__(self, connection):
        """ Parse the connection string.

            Initiate a trial connection with the database to check
            transactionality once instead of once per DB instance.

            Create database if option is enabled and database doesn't exist.
        """
        self.connection = connection
        DB = self.DB
        #
        db_flags = DB._parse_connection_string(connection, self.use_unicode)
        self._db_flags = db_flags

        # connect to server to determin tranasactional capabilities
        # can't use DB instance as it requires this information to work
        try:
            connection = MySQLdb.connect(**db_flags['kw_args'])
        except OperationalError:
            if self._create_db:
                kw_args = db_flags.get('kw_args',{}).copy()
                db = kw_args.pop('db',None)
                if not db: raise
                connection = MySQLdb.connect(**kw_args)
                create_query = "create database %s" % db
                if self.use_unicode:
                    create_query += (" default character set %s"
                            % DB.unicode_charset)
                connection.query(create_query)
                connection.store_result()
            else:
                raise
        transactional = connection.server_capabilities & CLIENT.TRANSACTIONS
        connection.close()

        # Some tweaks to transaction/locking db_flags based on server setup
        if db_flags['try_transactions'] == '-':
            transactional = False
        elif not transactional and db_flags['try_transactions'] == '+':
            raise NotSupportedError, "transactions not supported by this server"
        db_flags['transactions'] = transactional
        del db_flags['try_transactions']
        if transactional or db_flags['mysql_lock']:
            db_flags['use_TM'] = True

        # will not be 100% accurate in regard to per thread connections
        # but as close as we're going to get it.
        self.connected_timestamp = DateTime()

        # return self as the database connection object 
        # (assigned to _v_database_connection)
        return self

    def closeConnection(self):
        """ Close this threads connection. Used when DA is being reused
            but the connection string has changed. Need to close the DB
            instances and recreate to the new database. Only have to worry
            about this thread as when each thread hits the new connection
            string in the DA this method will be called.
        """
        ident = get_ident()
        try:
            self._pool_del(ident)
        except KeyError:
            pass

    def close(self):
        """ Used when manually closing the database. Resetting the pool
            dereferences the DB instances where they are then collected
            and closed.
        """
        self._db_pool = {}

    def _pool_set(self, key, value):
        """ Add a db to pool.
        """
        self._db_lock.acquire()
        try:
            self._db_pool[key] = value
        finally:
            self._db_lock.release()

    def _pool_get(self, key):
        """ Get db from pool. Read only, no lock necessary.
        """
        return self._db_pool.get(key)

    def _pool_del(self, key):
        """ Remove db from pool.
        """
        self._db_lock.acquire()
        try:
            del self._db_pool[key]
        finally:
            self._db_lock.release()

    def name(self):
        """ Return name of database connected to.
        """
        return self._db_flags['kw_args']['db']

    # Passthrough aliases for methods on DB class.
    def variables(self, *args, **kw):
        return self._access_db(method_id='variables', args=args, kw=kw)

    def tables(self, *args, **kw):
        return self._access_db(method_id='tables', args=args, kw=kw)

    def columns(self, *args, **kw):
        return self._access_db(method_id='columns', args=args, kw=kw)

    def query(self, *args, **kw):
        return self._access_db(method_id='query', args=args, kw=kw)

    def string_literal(self, *args, **kw):
        return self._access_db(method_id='string_literal', args=args, kw=kw)

    def unicode_literal(self, *args, **kw):
        return self._access_db(method_id='unicode_literal', args=args, kw=kw)

    def _access_db(self, method_id, args, kw):
        """
          Generic method to call pooled objects' methods.
          When the current thread had never issued any call, create a DB
          instance.
        """
        ident = get_ident()
        db = self._pool_get(ident)
        if db is None:
            db = self.DB(**self._db_flags)
            self._pool_set(ident, db)
        return getattr(db, method_id)(*args, **kw)


class DB(joinTM):

    Database_Error=_mysql.Error

    defs={
        FIELD_TYPE.CHAR: "i", FIELD_TYPE.DATE: "d", FIELD_TYPE.DATETIME: "d",
        FIELD_TYPE.DECIMAL: "n", FIELD_TYPE.NEWDECIMAL: "n",
        FIELD_TYPE.DOUBLE: "n", FIELD_TYPE.FLOAT: "n", FIELD_TYPE.INT24: "i",
        FIELD_TYPE.LONG: "i", FIELD_TYPE.LONGLONG: "l",
        FIELD_TYPE.SHORT: "i", FIELD_TYPE.TIMESTAMP: "d",
        FIELD_TYPE.TINY: "i", FIELD_TYPE.YEAR: "i",
        }

    conv=conversions.copy()
    conv[FIELD_TYPE.LONG] = int_or_long
    conv[FIELD_TYPE.DATETIME] = DateTime_or_None
    conv[FIELD_TYPE.DATE] = DateTime_or_None
    conv[FIELD_TYPE.DECIMAL] = float
    conv[FIELD_TYPE.NEWDECIMAL] = float
    del conv[FIELD_TYPE.TIME]
    # compatibility/fix for older mysqldb versions
    # without this you can get character arrays for some values
    # eg. group_concat()
    if MySQLdb_version < (1, 2, 2):
        conv[FIELD_TYPE.BLOB][0] = (FLAG.BINARY, str)

    _p_oid=_p_changed=_registered=None

    unicode_charset    = 'utf8' # hardcoded for now

    def __init__(self, connection=None, kw_args=None, use_TM=None,
            mysql_lock=None, transactions=None):
        self.connection     = connection # backwards compat
        self._kw_args       = kw_args
        self._mysql_lock    = mysql_lock
        self._use_TM        = use_TM
        self._transactions  = transactions
        self._forceReconnection()

    def close(self):
        """ Close connection and dereference.
        """
        if getattr(self,'db',None):
            self.db.close()
            self.db = None
    __del__ = close

    def _forceReconnection(self):
        """ (Re)Connect to database.
        """
        try: # try to clean up first
            self.db.close()
        except: pass
        self.db = MySQLdb.connect(**self._kw_args)
        # Newer mysqldb requires ping argument to attmept a reconnect.
        # This setting is persistent, so only needed once per connection.
        if MySQLdb_version >= (1, 2, 2):
            self.db.ping(True)

    @classmethod
    def _parse_connection_string(cls, connection, use_unicode=False):
        """ Done as a class method to both allow access to class attribute
            conv (conversion) settings while allowing for wrapping pool class
            use of this method. The former is important to allow for subclasses
            to override the conv settings while the latter is important so
            the connection string doesn't have to be parsed for each instance
            in the pool.
        """
        kw_args = {'conv':cls.conv}
        flags = {'kw_args':kw_args, 'connection':connection}
        if use_unicode:
            kw_args['use_unicode'] = use_unicode
            kw_args['charset']     = cls.unicode_charset
        items = connection.split()
        flags['use_TM'] = None
        if _mysql.get_client_info()[0] >= '5':
            kw_args['client_flag'] = CLIENT.MULTI_STATEMENTS
        if items:
            lockreq, items = items[0], items[1:]
            if lockreq[0] == "*":
                flags['mysql_lock'] = lockreq[1:]
                db_host, items = items[0], items[1:]
                flags['use_TM'] = True # redundant. eliminate?
            else:
                flags['mysql_lock'] = None
                db_host = lockreq
            if '@' in db_host:
                db, host = db_host.split('@',1)
                kw_args['db'] = db
                if ':' in host:
                    host, port = host.split(':',1)
                    kw_args['port'] = int(port)
                kw_args['host'] = host
            else:
                kw_args['db'] = db_host
            if kw_args['db'] and kw_args['db'][0] in ('+', '-'):
                flags['try_transactions'] = kw_args['db'][0]
                kw_args['db'] = kw_args['db'][1:]
            else:
                flags['try_transactions'] = None
            if not kw_args['db']:
                del kw_args['db']
            if items:
                kw_args['user'], items = items[0], items[1:]
            if items:
                kw_args['passwd'], items = items[0], items[1:]
            if items:
                kw_args['unix_socket'], items = items[0], items[1:]

        return flags

    def tables(self, rdb=0,
               _care=('TABLE', 'VIEW')):
        """ Returns list of tables.
        """
        r=[]
        a=r.append
        result = self._query("SHOW TABLES")
        row = result.fetch_row(1)
        while row:
            table_name = row[0][0]
            a({'table_name': table_name, 'table_type': 'table'})
            row = result.fetch_row(1)
        return r

    def columns(self, table_name):
        """ Returns list of columns for [table_name].
        """
        try:
            # Field, Type, Null, Key, Default, Extra
            c = self._query('SHOW COLUMNS FROM %s' % table_name)
        except:
            return ()
        r=[]
        for Field, Type, Null, Key, Default, Extra in c.fetch_row(0):
            info = {}
            field_default = ''
            if Default is not None:
                info['default'] = Default
                field_default = "DEFAULT '%s'" % Default
            if '(' in Type:
                end = Type.rfind(')')
                short_type, size = Type[:end].split('(',1)
                if short_type not in ('set','enum'):
                    if ',' in size:
                        info['scale'], info['precision'] = \
                                       map(int, size.split(',',1))
                    else:
                        info['scale'] = int(size)
            else:
                short_type = Type
            if short_type in field_icons:
                info['icon'] = short_type
            else:
                info['icon'] = icon_xlate.get(short_type, "what")
            info['name'] = Field
            info['type'] = short_type
            info['extra'] = Extra,
            info['description'] = ' '.join([Type, field_default, Extra or '',
                                        key_types.get(Key, Key or ''),
                                        Null == 'NO' and 'NOT NULL' or ''])
            info['nullable'] = (Null == 'YES') and 1 or 0
            if Key:
                info['index'] = True
                info['key'] = Key
            if Key == 'PRI':
                info['primary_key'] = True
                info['unique'] = True
            elif Key == 'UNI':
                info['unique'] = True
            r.append(info)
        return r

    def variables(self):
        """ Return dictionary of current mysql variable/values.
        """
        # variable_name, value
        variables = self._query('SHOW VARIABLES')
        return dict((name, value) for name, value in variables.fetch_row(0))

    def _query(self, query, force_reconnect=False):
        """
          Send a query to MySQL server.
          It reconnects automaticaly if needed and the following conditions are
          met:
           - It has not just tried to reconnect (ie, this function will not
             attempt to connect twice per call).
           - This conection is not transactionnal and has set no MySQL locks,
             because they are bound to the connection. This check can be
             overridden by passing force_reconnect with True value.
        """
        try:
            self.db.query(query)
        except OperationalError, m:
            if m[0] in query_syntax_error:
                raise OperationalError(m[0], '%s: %s' % (m[1], query))
            if ((not force_reconnect) and \
                    (self._mysql_lock or self._transactions)) or \
                    m[0] not in hosed_connection:
                LOG.warning('query failed: %s' % (query,))
                raise
            # Hm. maybe the db is hosed.  Let's restart it.
            if m[0] in hosed_connection:
                LOG.error('%s Forcing a reconnect.' % hosed_connection[m[0]])
            self._forceReconnection()
            self.db.query(query)
        except ProgrammingError, m:
            if m[0] in hosed_connection:
                self._forceReconnection()
                LOG.error('%s Forcing a reconnect.' % hosed_connection[m[0]])
            else:
                LOG.warning('query failed: %s' % (query,))
            raise
        return self.db.store_result()

    def query(self, query_string, max_rows=1000):
        """ API method for making the query to mysql.
        """
        self._use_TM and self._register()
        desc=None
        result=()
        for qs in filter(None, [q.strip() for q in query_string.split('\0')]):
            qtype = qs.split(None, 1)[0].upper()
            if qtype == "SELECT" and max_rows:
                qs = "%s LIMIT %d" % (qs,max_rows)
                r=0
            c = self._query(qs)
            if desc is not None:
                if c and (c.describe() != desc):
                    raise ProgrammingError(
                        'Multiple select schema are not allowed'
                        )
            if c:
                desc=c.describe()
                result=c.fetch_row(max_rows)
            else:
                desc=None

            if qtype == "CALL":
                # For stored procedures, skip the status result
                self.db.next_result()

        if desc is None: return (),()

        items=[]
        func=items.append
        defs=self.defs
        for d in desc:
            item={'name': d[0],
                  'type': defs.get(d[1],"t"),
                  'width': d[2],
                  'null': d[6]
                 }
            func(item)
        return items, result

    def string_literal(self, s):
        """ Called from zope to quote/escape strings for inclusion
            in a query.
        """
        return self.db.string_literal(s)

    def unicode_literal(self, s):
        """ Similar to string_literal but encodes it first.
        """
        return self.db.unicode_literal(s)

    # Zope 2-phase transaction handling methods
    def _begin(self, *ignored):
        """ Called from _register() upon first query.
        """
        try:
            self._transaction_begun = True
            self.db.ping()
            if self._transactions:
                self._query("BEGIN")
            if self._mysql_lock:
                self._query("SELECT GET_LOCK('%s',0)" % self._mysql_lock)
        except:
            LOG.error("exception during _begin", exc_info=True)
            raise ConflictError

    def _finish(self, *ignored):
        """ Called as final event in transaction system.
        """
        if not self._transaction_begun:
            return
        self._transaction_begun = False
        try:
            if self._mysql_lock:
                self._query("SELECT RELEASE_LOCK('%s')" % self._mysql_lock)
            if self._transactions:
                self._query("COMMIT")
        except:
            LOG.error("exception during _finish", exc_info=True)
            raise ConflictError

    def _abort(self, *ignored):
        """ Called in case of transactional error.
        """
        if not self._transaction_begun:
            return
        self._transaction_begun = False
        if self._mysql_lock:
            self._query("SELECT RELEASE_LOCK('%s')" % self._mysql_lock)
        if self._transactions:
            self._query("ROLLBACK")
        else:
            LOG.error("aborting when non-transactional")

    def _mysql_version(self):
        """ Return mysql server version.
            Note instances of this class are not persistent.
        """
        _version = getattr(self, '_version', None)
        if not _version:
            self._version = _version = self.variables().get('version')
        return _version

    def savepoint(self):
        """ Basic savepoint support.

            Raise AttributeErrors to trigger optimistic savepoint handling
            in zope's transaction code.
        """
        if self._mysql_version() < '5.0.2':
            # mysql supports savepoints in versions 5.0.3+
            LOG.warning("Savepoints unsupported with Mysql < 5.0.3")
            raise AttributeError

        if not self._transaction_begun:
            LOG.error("Savepoint used outside of transaction.")
            raise AttributeError

        return _SavePoint(self)


class _SavePoint(object):
    """ Simple savepoint object
    """
    def __init__(self, dm):
        self.dm = dm
        self.ident = ident = str(time.time()).replace('.','sp')
        dm._query("SAVEPOINT %s" % ident)

    def rollback(self):
        self.dm._query("ROLLBACK TO %s" % self.ident)





##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
__doc__='''Database Connection

$Id: DABase.py 522 2008-02-08 04:15:23Z eikenberry $'''
__version__='$Revision: 522 $'[11:-2]

import Shared.DC.ZRDB.Connection, sys
from Globals import HTMLFile
from ExtensionClass import Base
import Acquisition

class Connection(Shared.DC.ZRDB.Connection.Connection):
    _isAnSQLConnection=1

    manage_options=Shared.DC.ZRDB.Connection.Connection.manage_options+(
        {'label': 'Browse', 'action':'manage_browse'},
        )

    manage_browse=HTMLFile('browse',globals())

    info=None
        
    def tpValues(self):
        r=[]
        try:
            c = self._v_database_connection
        except AttributeError:
            self.connect(self.connection_string)
            c = self._v_database_connection
        for d in c.tables(rdb=0):
            try:
                name = d['table_name']
                b = TableBrowser()
                b.__name__ = name
                b._d = d
                b._c = c
                b.icon = table_icons.get(d['table_type'],'text')
                r.append(b)
            except:
                pass
        return r


class Browser(Base):
    def __getattr__(self, name):
        try: return self._d[name]
        except KeyError: raise AttributeError, name

class values:

    def len(self): return 1

    def __getitem__(self, i):
        try: return self._d[i]
        except AttributeError: pass
        self._d = self._f()
        return self._d[i]

class TableBrowser(Browser, Acquisition.Implicit):
    icon='what'
    description=check=''
    info=HTMLFile('table_info',globals())

    def tpValues(self):
        v=values()
        v._f=self.tpValues_
        return v

    def tpValues_(self):
        r=[]
        tname=self.__name__
        for d in self._c.columns(tname):
            b=ColumnBrowser()
            b._d=d
            b.icon=d['icon']
            b.table_name=tname
            r.append(b)
        return r
            
    def tpId(self): return self._d['table_name']
    def tpURL(self): return "Table/%s" % self._d['table_name']
    def name(self): return self._d['table_name']
    def type(self): return self._d['table_type']

class ColumnBrowser(Browser):
    icon='field'

    def check(self):
        return ('\t<input type=checkbox name="%s.%s">' %
                (self.table_name, self._d['name']))
    def tpId(self): return self._d['name']
    def tpURL(self): return "Column/%s" % self._d['name']
    def description(self): return " %s" % self._d['description']

table_icons={
    'table': 'table',
    'view':'view',
    'system_table': 'stable',
    }




def manage_addZMySQLConnection(self, id, title,
                                connection_string,
                                check=None, REQUEST=None):
    """Add a MySQL connection to a folder.

    Arguments:

        REQUEST -- The current request

        title -- The title of the ZMySQLDA Connection (string)

        id -- The id of the ZMySQLDA Connection (string)

        connection_string -- The connection string is of the form:

        '[*lock] [+/-][database][@host[:port]] [user [password [unix_socket]]]'

        or typically:

                'database user password'

        to use a MySQL server on localhost via the standard UNIX
        socket.  Only specify host if the server is on a remote
        system. You can use a non-standard port, if necessary. If the
        UNIX socket is in a non-standard location, you can specify the
        full path to it after the password. Hint: To use a
        non-standard port on the local system, use 127.0.0.1 for the
        host instead of localhost.
        
        Either a database or a host or both must be specified.     

        A '-' in front of the database tells ZMySQLDA to not use
        Zope's Transaction Manager, even if the server supports
        transactions. A '+' in front of the database tells ZMySQLDA
        that it must use transactions; an exception will be raised if
        they are not supported by the server. If neither '-' or '+'
        are present, then transactions will be enabled if the server
        supports them.  If you are using non-transaction safe tables
        (TSTs) on a server that supports TSTs, use '-'. If you require
        transactions, use '+'. If you aren't sure, don't use either.

        *lock at the begining of the connection string means to
        psuedo-transactional. When the transaction begins, it will
        acquire a lock on the server named lock (i.e. MYLOCK). When
        the transaction commits, the lock will be released. If the
        transaction is aborted and restarted, which can happen due to
        a ConflictError, you'll get an error in the logs, and
        inconsistent data. In this respect, it's equivalent to
        transactions turned off.

        Transactions are highly recommended. Using a named lock in
        conjunctions with transactions is probably pointless.

        """

class Connection:
    """MySQL Connection Object"""

    __constructor__ = manage_addZMySQLConnection





class DB:

    """This is the ZMySQLDA Database Connection Object."""

    def __init__(self,connection):
        """
            connection
               blah blah
        """
        
    def tables(self, rdb=0,
               _care=('TABLE', 'VIEW')):
        """Returns a list of tables in the current database."""

    def columns(self, table_name):
        """Returns a list of column descriptions for 'table_name'."""

    def query(self,query_string, max_rows=1000):
        """Execute 'query_string' and return at most 'max_rows'."""

    def _begin(self, *ignored):
        """Begin a transaction (when TM is enabled)."""
        
    def _finish(self, *ignored):
        """Commit a transaction (when TM is enabled)."""

    def _abort(self, *ignored):
        """Rollback a transaction (when TM is enabled)."""



##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
__doc__='''Generic Database Adapter Package Registration

$Id: __init__.py 542 2008-08-14 21:40:36Z eikenberry $'''
__version__='$Revision: 542 $'[11:-2]

import DA

misc_=DA.misc_

def initialize(context):

    context.registerClass(
        DA.Connection,
        permission='Add Z MySQL Database Connections',
        constructors=(DA.manage_addZMySQLConnectionForm,
                      DA.manage_addZMySQLConnection),
    )

    context.registerHelp()
    context.registerHelpTitle('ZMySQLDA')




#!/usr/bin/python
from setuptools import setup, find_packages

# See MANIFEST.in for filelist

setup(
    name="Products.ZMySQLDA",
    version="3.1.1",
    license="Zope Public License (ZPL) Version 1.0",
    author="John Eikenberry",
    author_email="jae@zhar.net",
    url="http://sourceforge.net/projects/mysql-python",
    description="MySQL Zope2 adapter.",
    long_description=("MySQL Database Adapter for Zope 2. Extensively "
        "reworked for stability and compatibility with versions 2.8+ and "
        "modern MySQL versions. New features from auto-creating database "
        "to limited Unicode support."),
    packages=find_packages(),
    include_package_data=True,
    namespace_packages=['Products'],
    zip_safe=False,
    install_requires = "MySQL-python >= 1.2.1",
    )




##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
database_type='MySQL'
__doc__='''%s Database Connection

$Id: DA.py 648 2010-11-10 02:06:44Z eikenberry $''' % database_type
__version__='$Revision: 648 $'[11:-2]

import os, os.path
from db import DBPool, DB
from thread import allocate_lock

from Shared.DC import ZRDB
import DABase
from App.ImageFile import ImageFile
import Globals

manage_addZMySQLConnectionForm=Globals.HTMLFile('connectionAdd',globals())

def manage_addZMySQLConnection(self, id, title,
                                connection_string,
                                check=None,
                                use_unicode=None,
                                auto_create_db=None,
                                REQUEST=None):
    """Add a DB connection to a folder"""
    self._setObject(id,
            Connection(id, title, connection_string, check,
                use_unicode=use_unicode,
                auto_create_db=auto_create_db))
    if REQUEST is not None: return self.manage_main(self,REQUEST)

# Connection Pool for connections to MySQL.
# Maps one mysql client connection to one DA object instance.
database_connection_pool_lock = allocate_lock()
database_connection_pool = {}

# Combining connection pool with the DB pool gets you
# one DA/connection per connection with 1 DBPool with 1 DB/thread
# pool_id -> DA -> DBPool -> thread id -> DB
# dc_pool[pool_id] == DBPool_instance
# DBPool_instance[thread id] == DB instance

class Connection(DABase.Connection):
    """ ZMySQL Database Adapter Connection.
    """
    database_type=database_type
    id='%s_database_connection' % database_type
    meta_type=title='Z %s Database Connection' % database_type
    icon='misc_/Z%sDA/conn' % database_type

    auto_create_db = True
    use_unicode = False
    _v_connected = ''

    manage_properties=Globals.HTMLFile('connectionEdit', globals())

    def factory(self):
        """ Base API. Returns factory method for DB connections.
        """
        return DB
    
    def _pool_key(self):
        """ Return key used for DA pool.
        """
        return self.getPhysicalPath()

    def connect(self, s):
        """ Base API. Opens connection to mysql. Raises if problems.
        """
        pool_key = self._pool_key()
        connection = database_connection_pool.get(pool_key)
        if connection is not None and connection.connection == s:
            self._v_database_connection = connection
            self._v_connected = connection.connected_timestamp
        else:
            if connection is not None:
                connection.closeConnection()
            DB = self.factory()
            DB = DBPool(DB, create_db=self.auto_create_db,
                    use_unicode=self.use_unicode)
            database_connection_pool_lock.acquire()
            try:
                database_connection_pool[pool_key] = connection = DB(s)
            finally:
                database_connection_pool_lock.release()
            self._v_database_connection = connection
            # XXX If date is used as such, it can be wrong because an 
            # existing connection may be reused. But this is suposedly
            # only used as a marker to know if connection was successfull.
            self._v_connected = connection.connected_timestamp

        return self # ??? why doesn't this return the connection ???
        
    def sql_quote__(self, v, escapes={}):
        """ Base API. Used to message strings for use in queries.
        """
        try:
            connection = self._v_database_connection
        except AttributeError:
            self.connect(self.connection_string)
            connection = self._v_database_connection
        if self.use_unicode and isinstance(v, unicode):
            return connection.unicode_literal(v)
        else:
            return connection.string_literal(v)

    def __init__(self, id, title, connection_string, check,
                    use_unicode=None,
                    auto_create_db=None):
        """ Instance setup. Optionally opens the connection (check arg).
        """
        if use_unicode is not None:
            self.use_unicode = bool(use_unicode)
        if auto_create_db is not None:
            self.auto_create_db = bool(auto_create_db)
        return DABase.Connection.__init__(self, id, title, connection_string,
            check)

    def __setstate__(self,state):
        """ Skip super's __setstate__ as it connects which we don't want
            due to pool_key depending on acquisition.
        """
        Globals.Persistent.__setstate__(self, state)

    def manage_edit(self, title, connection_string,
                    check=None,
                    use_unicode=None,
                    auto_create_db=None):
        """ Zope management API.
        """
        if use_unicode is not None:
            self.use_unicode = bool(use_unicode)
        if auto_create_db is not None:
            self.auto_create_db = bool(auto_create_db)
        return DABase.Connection.manage_edit(self, title, connection_string,
                    check=None)


classes=('DA.Connection',)

meta_types=(
    {'name':'Z %s Database Connection' % database_type,
     'action':'manage_addZ%sConnectionForm' % database_type,
     },
    )

folder_methods={
    'manage_addZMySQLConnection':
    manage_addZMySQLConnection,
    'manage_addZMySQLConnectionForm':
    manage_addZMySQLConnectionForm,
    }

__ac_permissions__=(
    ('Add Z MySQL Database Connections',
     ('manage_addZMySQLConnectionForm',
      'manage_addZMySQLConnection')),
    )

misc_={'conn': ImageFile(
    os.path.join(
        os.path.dirname(ZRDB.__file__), 'www', 'DBAdapterFolder_icon.gif')
    )}

for icon in ('table', 'view', 'stable', 'what',
        'field', 'text','bin','int','float',
        'date','time','datetime'):
    misc_[icon]=ImageFile(os.path.join('icons','%s.gif') % icon, globals())

